import 'package:bizdoc/core/app.dart';
import 'package:dio/dio.dart';
import 'package:location/location.dart';
import 'dart:math';
import 'package:bizdoc/services/base.dart';

const String googleMapsBaseUri = "https://maps.googleapis.com/maps/api";

class GeoService extends Service {
  String? _apiKey;

  Future<Iterable<Address>> addresses(String name,
      {String? countryCode}) async {
    final queryParameters = {'input': name};
    if (countryCode != null) {
      queryParameters['components'] = 'country:$countryCode';
    }
    final data =
        await _get('place/autocomplete/json', queryParameters: queryParameters);
    return (data['predictions'] as List).map((e) => Address.fromJson(e));
  }

  Future<Iterable<BusinessInfo>> nearby(
      {String? name,
      required double latitude,
      required double longitude,
      int radius = 500}) async {
    final queryParameters = {
      'location': '$latitude,$longitude',
      'radius': radius,
      // 'language': window.locale
    };
    if (name != null) queryParameters['name'] = name;
    final data =
        await _get('place/nearbysearch/json', queryParameters: queryParameters);
    return (data['results'] as List).map((e) => BusinessInfo.fromJson(e));
  }

  Future<LatLng> geocode(String placeId) async {
    final data =
        await _get('geocode/json', queryParameters: {'place_id': placeId});
    final results = data['results'] as Iterable;
    final location = results.elementAt(0)['geometry']['location'];
    return LatLng.fromJson(location);
  }

  Future<Map> _get(String uri,
      {required Map<String, dynamic> queryParameters}) async {
    queryParameters['key'] = _apiKey;
    final response = await Dio()
        .get('$googleMapsBaseUri/$uri', queryParameters: queryParameters);
    var data = response.data;
    if (data['status'] != 'OK') throw data['status'];
    return data;
  }

  Future<LatLng> address(
      {required double latitude, required double longitude}) async {
    final data = await _get('geocode/json',
        queryParameters: {'latlng': '$latitude,$longitude'});
    final results = data['results'] as Iterable;
    final location = results.elementAt(0)['geometry']['location'];
    return LatLng.fromJson(location);
  }

  void initialize(AppState appState) {
    _apiKey = appState.googleMaps!.apiKey;
  }
}

final geo = GeoService();

class Address {
  final String address;
  final String placeId;
  final Iterable<String>? types;
  Address.fromJson(Map<String, dynamic> json)
      : this(json['description'], json['place_id'], List.from(json['types']));

  Address(this.address, this.placeId, this.types);
}

class BusinessInfo {
  final String name;
  final String placeId;
  final Iterable<String>? types;
  final LatLng location;
  final String? address;
  BusinessInfo.fromJson(Map<String, dynamic> json)
      : this(json['name'], json['place_id'], List.from(json['types']),
            LatLng.fromJson(json['geometry']['location']), json['vicinity']);

  BusinessInfo(
      this.name, this.placeId, this.types, this.location, this.address);
}

class LatLng {
  final Map<String, dynamic> _json;

  double get latitude => _json['lat'];
  double get longitude => _json['lng'];

  Map<String, dynamic> asJson() => _json;

  LatLng(this._json);
  LatLng.fromJson(Map<String, dynamic> json) : this(json);
  LatLng.from(LocationData data)
      : _json = {'lat': data.latitude, 'lng': data.longitude};
}

const earthRadius = 6371;

_degreesToRadians(degrees) => degrees * (pi / 180);
_radiansToDegrees(radians) => radians * (180 / pi);
_greatCircleDistance(angle) => 2 * pi * earthRadius * (angle / 360);

distance(LatLng locationX, LatLng locationY) {
  final locationXLatRadians = _degreesToRadians(locationX.latitude);
  final locationYLatRadians = _degreesToRadians(locationY.latitude);
  final centralSubtendedAngle = _radiansToDegrees(acos(
      sin(locationXLatRadians) * sin(locationYLatRadians) +
          cos(locationXLatRadians) *
              cos(locationYLatRadians) *
              cos(_degreesToRadians(locationX.longitude > locationY.longitude
                  ? (locationX.longitude - locationY.longitude)
                  : locationY.longitude - locationX.longitude))));
  return _greatCircleDistance(centralSubtendedAngle);
}
